﻿/************************************************************
 *	File:		MasterCAB.cs
 *
 * 	Author:		Samuel Joski
 *	
 *	Contents:	Partial Class MasterCAB:Form
 *				Class GLOBALS
 * 	
 * 	Purpose:	The GLOBALS class is used to represent the
 * 				meaning of railroad resource various statuses
 * 				and directions. It is identical to the GLOBALS
 * 				class found in CCRServer.cs
 * 				
 * 				The MasterCAB class is responsible for handlers
 * 				for events on the MasterCAB form, displaying
 * 				appropriate railroad resources to the form,
 * 				as well as reading and writing to the files
 * 				for trains and turnouts.
 * *********************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Timers;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace MasterCAB
{
	public partial class MasterCAB : Form
	{
		//Class variables
		private int NUM_TURNOUTS;		//Stores the number of turnouts read from the turnout file
		private int NUM_TRAINS;			//Stores the number of trains read from the train file
		private int SELECTED_TRAIN;		//Used for quick index access for the CAB's selected train

		//Class Data Structures
		private Train[] trains;					//Stores the information about trains from the trains file	
		private Turnout[] turnouts;             //Stores the information about turnouts from the turnouts file

		//Programmatically defined turnout checkboxes
		private CheckBox[] turnoutCBs;

		//fileRead Timer
		private System.Timers.Timer fileRead;

		public MasterCAB()
		{
			InitializeComponent();
			SetupTrains();
			SetupTurnouts();

			SELECTED_TRAIN = GLOBALS.UNSELECTED;

			//Setup fileRead timer
			fileRead = new System.Timers.Timer();
			fileRead.Elapsed += new ElapsedEventHandler(readFilesEvent);
			fileRead.Interval = 250;
			fileRead.Enabled = true;
		}

		//Timer ElapsedEventHandler Functions
		private void readFilesEvent(object source, ElapsedEventArgs e)
		{
			/**************************************************************
			 *	Function Name:		readFilesEvent
			 *	Input Parameters:	object source, ElapsedEventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is the event triggered by the 
			 *				fileReader timer, which is triggered every 100ms.
			 *				
			 *				The functions readTrains() and readTurnouts() are
			 *				called from this event to update the structures if
			 *				they contain new values, which occurs when the
			 *				server has modified the values.
			 * ************************************************************/

			//Check the turnouts
			System.Threading.Monitor.Enter(turnouts);
			readTurnouts();
			System.Threading.Monitor.Exit(turnouts);

			//Check the trains
			System.Threading.Monitor.Enter(trains);
			readTrains();

			//Check that the server did not reject an occupy request
			if(SELECTED_TRAIN != GLOBALS.UNSELECTED)
			{
				if(!trains[SELECTED_TRAIN].isOccupied())
				{
					MessageBox.Show("Sorry, the selected train is currently occupied.", "Train selection failed");
					SELECTED_TRAIN = GLOBALS.UNSELECTED;
				}
			}
			System.Threading.Monitor.Exit(trains);

			updateDisplays();
		}

		//Setup Data Structures Functions
		private void SetupTrains()
		{
			/**************************************************************
			 *	Function Name:		SetupTrains
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function does the following:
			 *					* Reads the number of trains in the file
			 *						"trains.txt" via getNumObjects(). The
			 *						number is stored as the class integer
			 *						NUM_TRAINS.
			 *					* Creates int NUM_TRAINS many trains in
			 *						the trains array.
			 *					* Reads in the data from "trains.txt" to
			 *						fill the trains in the trains array
			 *					* Adds NUM_TRAINS many items to the ListBox
			 *						lbTrains on the Master CAB form, which
			 *						will be used for selecting a train to
			 *						be controlled.
			 * ************************************************************/

			//Set NUM_TRAINS to number of objects in "trains.txt"
			NUM_TRAINS = getNumObjects("trains.txt");
			trains = new Train[NUM_TRAINS];

			readTrains();

			//Add the trains to the ListBox lbTrains
			for (int i = 0; i < NUM_TRAINS; i++)
			{
				if (trains[i] != null)
				{
					lbTrains.Items.Add(trains[i].Address());
				}
			}
		}
		private void SetupTurnouts()
		{
			/**************************************************************
			 *	Function Name:		SetupTurnouts
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function does the following:
			 *					* Reads the number of turnouts in the file
			 *						"turnouts.txt" via getNumObjects(). The
			 *						number is stored as the class integer
			 *						NUM_TURNOUTS.
			 *					* Creates int NUM_TURNOUTS many turnouts in
			 *						the turnouts array.
			 *					* Reads in the data from "turnouts.txt" to
			 *						fill the turnouts in the turnouts array
			 *					* Creates NUM_TURNOUTS many checkboxes for
			 *						the Master CAB form that will act as
			 *						toggles for the turnouts.
			 * ************************************************************/

			//Set NUM_TURNOUTS to number of objects in "turnouts.txt"
			NUM_TURNOUTS = getNumObjects("turnouts.txt");

			//Create the turnouts
			turnouts = new Turnout[NUM_TURNOUTS];
			for (int i = 0; i < NUM_TURNOUTS; i++)
			{
				turnouts[i] = new Turnout(i);
			}
			readTurnouts();

			//Create the turnout checkboxes
			turnoutCBs = new CheckBox[NUM_TURNOUTS];
			for (int i = 0; i < NUM_TURNOUTS; i++)
			{
				//Create and Configure the checkbox
				turnoutCBs[i] = new CheckBox();
				turnoutCBs[i].Appearance = Appearance.Button;
				turnoutCBs[i].AutoSize = true;
				turnoutCBs[i].TabIndex = 1;
				turnoutCBs[i].UseVisualStyleBackColor = true;

				//Place the checkbox based on the number that will eventually be placed
				int x = 5 + (turnoutCBs[i].Size.Width * (i / 3));
				int y = 15 + (turnoutCBs[i].Size.Height * (i % 3));
				turnoutCBs[i].Location = new Point(x, y);

				//Set checkbox toggle
				if (turnouts[i].Direction() == GLOBALS.BOTTOM)
				{
					turnoutCBs[i].Checked = true;
				}

				//Set text, name, and handler
				turnoutCBs[i].Name = "turnoutCB" + i;
				turnoutCBs[i].Text = "Turnout " + i;
				turnoutCBs[i].Tag = i;
				turnoutCBs[i].Click += new EventHandler(turnoutOnClick);
				gbTurnouts.Controls.Add(turnoutCBs[i]);
			}
			Controls.Add(gbTurnouts);
		}

		//Master CAB Form Functions
		private void updateDisplays()
		{
			/**************************************************************
			 *	Function Name:		updateDisplays
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function sets the appearance of several
			 *				elements on the MasterCAB form:
			 *					* ListBox		lbTrains
			 *					* TextBox		tbSpeed
			 *					* TextBox		tbDirection
			 *					* TextBox		tbStatus
			 *					* CheckBoxes	cbTurnouts
			 *				
			 *				ListBox lbTrains' content is set to match any
			 *				trains that may have been read in from the file
			 *				on the fileReader Timer.
			 *				
			 *				TextBox tbSpeed's text is set to the speed of
			 *				the train which is indexed as SELECTED_TRAIN 
			 *				in the trains array.
			 *				
			 *				TextBox tbDirection's text is set to the
			 *				forward direction of the train.
			 *				
			 *				Forward direction is determined by the train's
			 *				direction and orientation values:				
			 *					* FORWARD	if	direction == orientation
			 *					* BACKWARD	if	direction != orientation
			 *					
			 *				TextBox tbStatus is set to the status of the
			 *				train:
			 *					GLOBALS.HALTED		"TRAIN HALTED"
			 *					GLOBALS.DEAD_END	"AT DEAD END"
			 *					GLOBALS.MOVING		"TRAIN MOVING"
			 *					DEFAULT				"UNKNOWN"
			 *					
			 *				The CheckBoxes cbTurnouts are set to have their
			 *				checked appearance match their respective
			 *				Turnout data structure.
			 * ************************************************************/

			//MethodInvoker needed to access turnout objects
			BeginInvoke((MethodInvoker)delegate ()
			{
				//Look through the trains array
				for (int i = 0; i < NUM_TRAINS; i++)
				{
					//If the index is valid
					if (trains[i] != null)
					{
						//If the ListBox does not already contain it
						if (!lbTrains.Items.Contains(trains[i].Address()))
						{
							//Add it
							lbTrains.Items.Add(trains[i].Address());
						}
					}
				}

				//Only update the train information if a train was selected
				if (SELECTED_TRAIN != -1)
				{
					//Set the TextBox tbSpeed to the train's speed
					tbSpeed.Text = trains[SELECTED_TRAIN].Speed().ToString();

					/*Set the TextBox tbDirection to the train's forward direction. The forward
						direction is FORWARD if the train's direction matches its orientation.*/
					if (trains[SELECTED_TRAIN].Direction() == trains[SELECTED_TRAIN].Orientation())
					{
						tbDirection.Text = "Forward";
					}
					else
					{
						tbDirection.Text = "Backward";
					}

					/*Set the TextBox tbStatus to the train's status.*/
					switch (trains[SELECTED_TRAIN].Status())
					{
						case GLOBALS.HALTED:
							tbStatus.Text = "TRAIN HALTED";
							break;
						case GLOBALS.DEAD_END:
							tbStatus.Text = "AT DEAD END";
							break;
						case GLOBALS.MOVING:
							tbStatus.Text = "TRAIN MOVING";
							break;
						default:
							tbStatus.Text = "UNKNOWN";
							break;
					}
				}

				//Else no train is selected, remove information from the displays
				else
				{
					tbSpeed.Text = string.Empty;
					tbDirection.Text = string.Empty;
					tbStatus.Text = string.Empty;
					tbSelectedTrain.Text = string.Empty;
				}

				//Update turnout buttons
				for (int i = 0; i < NUM_TURNOUTS; i++)
				{
					//Set checkbox toggle
					if (turnouts[i].Direction() == GLOBALS.BOTTOM)
					{
						turnoutCBs[i].Checked = true;
					}
					else
					{
						turnoutCBs[i].Checked = false;
					}

					//Check if the turnout is occupied and disable the button if it is
					if (turnouts[i].Occupied())
					{
						turnoutCBs[i].Enabled = false;
					}
					else
					{
						turnoutCBs[i].Enabled = true;
					}
				}
			});
		}

		//Form Objects' onClick Functions
		private void turnoutOnClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		turnoutOnClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called when a turnout CheckBox
			 *				is clicked. This function updates the appropriate
			 *				turnout object in the turnouts array. The index
			 *				of the appropriate turnout object is determined
			 *				via the CheckBox Tag.
			 *				
			 *				writeTurnout() is then called to write to the
			 *				turnout file the updated turnout object.
			 *				
			 *	NOTES:		***SMJ*** THIS FUNCTION SHOULD ALSO CHECK FOR
			 *				FEEDBACK FROM THE SERVER IF THE TURNOUT REQUEST
			 *				WAS UNSUCCESSFUL.
			 * ************************************************************/
			System.Threading.Monitor.Enter(turnouts);
			CheckBox myCB = (CheckBox)sender;

			if (myCB.Checked)
			{
				turnouts[Convert.ToInt32(myCB.Tag)].setDirection(GLOBALS.BOTTOM);
			}
			else
			{
				turnouts[Convert.ToInt32(myCB.Tag)].setDirection(GLOBALS.TOP);
			}
			writeTurnout(turnouts[Convert.ToInt32(myCB.Tag)]);
			System.Threading.Monitor.Exit(turnouts);
		}
		private void speedOnClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		speedOnClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called when a speed Button is
			 *				clicked. This function updates the appropriate
			 *				train object in the trains array. The index of
			 *				the appropriate train object is determined via
			 *				the SELECTED_TRAIN integer.
			 *				
			 *				writeTrain() is then called to write to the
			 *				train file the updated train object.
			 *				
			 *				updateDisplays() is called to adjust displayed
			 *				values of speed and direction.
			 * ************************************************************/
			if (trainSelected())
			{
				Button myBTN = (Button)sender;
				if (myBTN.Tag.ToString() == "Positive")
				{
					trains[SELECTED_TRAIN].setSpeed(trains[SELECTED_TRAIN].Speed() + 1);
				}
				else if (myBTN.Tag.ToString() == "Negative")
				{
					trains[SELECTED_TRAIN].setSpeed(trains[SELECTED_TRAIN].Speed() - 1);
				}
				else
				{
					//DEFAULT NOTHING TO DO (Some other button assigned this function?)
				}
				writeTrain(trains[SELECTED_TRAIN]);
				updateDisplays();
			}
		}
		private void stopOnClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		stopOnClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called when the stop Button is
			 *				clicked. This function updates the appropriate
			 *				train object in the trains array to have a 
			 *				speed of 0. The index of the appropriate train
			 *				object is determined via the SELECTED_TRAIN
			 *				integer.
			 *				
			 *				writeTrain() is then called to write to the
			 *				train file the updated train object.
			 *				
			 *				updateDisplays() is called to adjust displayed
			 *				values of speed and direction.
			 * ************************************************************/
			if (trainSelected())
			{
				trains[SELECTED_TRAIN].setSpeed(GLOBALS.STOPPED_SPEED);
				writeTrain(trains[SELECTED_TRAIN]);
				updateDisplays();
			}
		}
		private void trainSelectOnClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		trainSelectedOnClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function sets the TextBox tbSelectedTrain's
			 *				text to the selected item in the ListBox
			 *				lbTrains if the corresponding train is not
			 *				occupied. If the MasterCAB is able to occupy the
			 *				train, then the previously selected train (if any)
			 *				is released, and the newly selected train is
			 *				occupied. 
			 *				
			 *				UpdateDisplays() is called to refresh
			 *				the information on screen to match the information
			 *				in the corresponding train that was selected.
			 *				
			 *				The class integer SELECTED_TRAIN is set to the
			 *				selected item for lookups that the interface
			 *				may perform.
			 * ************************************************************/
			System.Threading.Monitor.Enter(trains);

			//Make sure something was selected before the click
			if (lbTrains.SelectedItem != null)
			{
				string selected = lbTrains.SelectedItem.ToString();

				//Look through the trains array for the corresponding train
				for (int i = 0; i < trains.Length; i++)
				{
					//If the index is valid
					if(trains[i] != null)
					{
						//If the address of the index matches
						if(trains[i].Address() == Convert.ToInt32(selected))
						{
							//If the train is able to be occupied (not already occupied)
							if(trains[i].Occupy())
							{
								//Check if a train was previously selected
								if (tbSelectedTrain.Text != string.Empty)
								{
									//Loop through the trains array to look for the old train
									for (int j = 0; j < trains.Length; j++)
									{
										//If the index is valid
										if (trains[j] != null)
										{
											//If the address of the train matches the address of the previously occupied train
											if (trains[j].Address() == Convert.ToInt32(tbSelectedTrain.Text))
											{
												//Release the train from occupation
												trains[j].Release();
												writeTrain(trains[j]);
											}
										}
									}
								}
								/*Set the SelectedTrain text to the selected train, and write to the file
									that it is now occupied. Update displays*/
								tbSelectedTrain.Text = selected;
								SELECTED_TRAIN = trains[i].ID();
								writeTrain(trains[i]);
								updateDisplays();
							}
							//If unable to occupy, display that the train is in use.
							else
							{
								MessageBox.Show("The selected train is currently in use.", "Unable to select train");
							}
						}
					}
				}
			}
			System.Threading.Monitor.Exit(trains);
		}

		//File Read/Write Functions
		private void readTurnouts()
		{
			/**************************************************************
			 *	Function Name:		readTurnouts
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the "turnouts.txt" file
			 *				to input into the turnout array all turnouts that
			 *				are present in the file.
			 *				
			 *				The input parameters to the turnout class are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Direction		int (GLOBALS)	2
			 *					* Occupied		bool			3
			 *					
			 *				This function will continue to attempt to access
			 *				the file which may be locked by another process.
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("turnouts.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[NUM_TURNOUTS];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						//Read and split the line (comma delimited)
						lines[i] = sr.ReadLine();
						string[] lineinfo = lines[i].Split(',');

						//Convert file values
						int id = Convert.ToInt32(lineinfo[0]);								//Get the ID of the turnout
						int address = Convert.ToInt32(lineinfo[1]);							//Get the address of the turnout
						int direction = Convert.ToInt32(lineinfo[2]);						//Get the direction of the turnout
						bool occupied = Convert.ToBoolean(Convert.ToInt32(lineinfo[3]));    //Get the occupied bool of the turnout

						//Set train object to file values
						turnouts[id].setAddress(address);
						turnouts[id].setDirection(direction);
						turnouts[id].setOccupied(occupied);
					}
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
		}
		private void readTrains()
		{
			/**************************************************************
			 *	Function Name:		readTrains
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the "trains.txt" file
			 *				to input into the trains array all trains that
			 *				are present in the file.
			 *				
			 *				The input parameters to the train class are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Location		int				2
			 *					* Speed			int				3
			 *					* Direction		int (GLOBALS)	4
			 *					* Orientation	int (GLOBALS)	5
			 *					* Status		int (GLOBALS)	6
			 *					* Headlight		bool			7
			 *					* Horn			bool			8
			 *					* Occupied		bool			9
			 *					
			 *				This function will continue to attempt to access
			 *				the file which may be locked by another process.
			 * ************************************************************/
			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[NUM_TRAINS];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						//Read and split the line (comma delimited)
						lines[i] = sr.ReadLine();

						//If there is a train on the line
						if (lines[i] != null && lines[i] != string.Empty)
						{
							string[] lineinfo = lines[i].Split(',');

							//Convert file values
							int id = Convert.ToInt32(lineinfo[0]);                              //Get the ID of the train
							int address = Convert.ToInt32(lineinfo[1]);                         //Get the address of the train
							int location = Convert.ToInt32(lineinfo[2]);                        //Get the location of the train
							int speed = Convert.ToInt32(lineinfo[3]);                           //Get the speed of the train
							int direction = Convert.ToInt32(lineinfo[4]);                       //Get the direction of the train
							int orientation = Convert.ToInt32(lineinfo[5]);                     //Get the orientation of the train
							int status = Convert.ToInt32(lineinfo[6]);                          //Get the status of the train
							bool headlight = Convert.ToBoolean(Convert.ToInt32(lineinfo[7]));   //Get the headlight of the train
							bool horn = Convert.ToBoolean(Convert.ToInt32(lineinfo[8]));        //Get the horn of the train
							bool occupied = Convert.ToBoolean(Convert.ToInt32(lineinfo[9]));    //Get the occupied status of the train

							//Create the train if it did not exist
							if (trains[i] == null)
							{
								trains[i] = new Train(id);
							}

							//Set train object to file values
							trains[id].setLocation(location);
							trains[id].setAddress(address);
							trains[id].setSpeed(speed);
							trains[id].setDirection(direction);
							trains[id].setOrientation(orientation);
							trains[id].setStatus(status);
							trains[id].setHeadlight(headlight);
							trains[id].setHorn(horn);
							trains[id].setOccupied(occupied);
						}

						//Check if a train has been removed from the server
						else if (lines[i] == string.Empty && trains[i] != null)
						{
							/*Change selected train to UNSELECTED if the selected train was removed.
								This should only happen on the server shutdown.*/
							if(i == SELECTED_TRAIN)
							{
								SELECTED_TRAIN = GLOBALS.UNSELECTED;
							}

							//Remove train from display and set to null
							lbTrains.Items.Remove(trains[i].Address());
							trains[i] = null;
						}
					}
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
		}
		private void writeTurnout(Turnout turnout)
		{
			/**************************************************************
			 *	Function Name:		writeTurnout
			 *	Input Parameters:	Turnout turnout
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function writes to the "turnouts.txt" file
			 *				to output the turnout parameter. This function 
			 *				will continue to attempt to access the file
			 *				which may be locked by another process.
			 *				
			 *				The output parameters to the train file are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Direction		int (GLOBALS)	2
			 *					* Occupied		bool			3
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("turnouts.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[NUM_TURNOUTS];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();

						/*If the ID of the turnout in the file matches the parameter, then
							set the line to the InfoLine of the turnout.*/
						if (lines[i][0].ToString() == turnout.ID().ToString())
						{
							lines[i] = turnout.GetInfoLine();
						}
					}

					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
		}
		private void writeTrain(Train train)
		{
			/**************************************************************
			 *	Function Name:		writeTrain
			 *	Input Parameters:	Train train
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function writes to the "trains.txt" file
			 *				to output the train parameter. This function 
			 *				will continue to attempt to access the file
			 *				which may be locked by another process.
			 *				
			 *				The output parameters to the train file are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Location		Track			2
			 *					* Speed			int				3
			 *					* Direction		int (GLOBALS)	4
			 *					* Orientation	int (GLOBALS)	5
			 *					* Status		int (GLOBALS)	6
			 *					* Headlight		bool			7
			 *					* Horn			bool			8
			 *					* Occupied		bool			9
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[NUM_TRAINS];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();

						//If there is a train on the line
						if (lines[i] != null && lines[i] != string.Empty)
						{
							/*If the ID of the train in the file matches the parameter, then
								set the line to the InfoLine of the train.*/
							if (lines[i][0].ToString() == train.ID().ToString())
							{
								lines[i] = train.GetInfoLine();
							}
						}
					}

					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
		}
		private int getNumObjects(string filename)
		{
			/**************************************************************
			 *	Function Name:		getNumTrains
			 *	Input Parameters:	string filename
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the a file to count the
			 *				number of lines in the file. The string parameter
			 *				filename is passed to designate which object will
			 *				be read in.
			 *				
			 *					* Trains	- "trains.txt"
			 *					* Turnouts	- "turnouts.txt"
			 *				
			 *				This function will continue to attempt to access
			 *				the file (which may be locked by another process)
			 *				until it can successfully open and read from the
			 *				file.
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop
			int count = 0;			//Counter for the number of lines in the file (i.e. number of trains)

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					//Loop until the end of file to count each line
					while (!sr.EndOfStream)
					{
						sr.ReadLine();
						count++;
					}
					
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
			return count;
		}

		//Misc. Functions
		private bool trainSelected()
		{
			/**************************************************************
			 *	Function Name:		trainSelected
			 *	Input Parameters:	none
			 *	Output Parameters:	bool via return
			 *	
			 *	Purpose:	This function checks the textBox tbSelectedTrain
			 *				to verify that it has been populated by a user
			 *				selection from the ListBox lbTrains via the
			 *				function trainSelectedOnClick().
			 *				
			 *					* return TRUE if not empty string
			 *					* returns FALSE if empty string
			 * ************************************************************/
			return (tbSelectedTrain.Text.ToString() != string.Empty);
		}
	}
	class GLOBALS
	{
		//Directions and orientations
		public const int LEFT = 0;
		public const int RIGHT = 1;

		//Turnout states
		public const int TOP = 0;
		public const int BOTTOM = 1;

		//Track speed limit
		public const int DEFAULT_SPEED = 5;
		public const int STOPPED_SPEED = 0;

		//Train states
		public const int HALTED = 0;
		public const int DEAD_END = 1;
		public const int MOVING = 2;

		//Misc.
		public const int UNSELECTED = -1;
	}
}